﻿/*
 * Created by SharpDevelop.
 * User: Andreas
 * Date: 05.01.2018
 * Time: 00:05
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */

using System;
using CevoAILib;


namespace AI
{	
    /// <summary>
    /// Define Settler Jobs
    /// BuildImprovements = Build Irrigation and then Roads
    /// </summary>
    public enum SettlerToDo
    {
        None = 0x00, BuildIrrigation = 0x01, BuildRoads = 0x02, BuildImprovements = 0x03,
        SeekNewSettleLocation = 0x04, MoveToCity = 0x05
    }

    /// <summary>
    /// Define main movement direction for Settle Location Search
    /// </summary>
    public enum MapDirection
    {
        North = 0x01, NorthEast = 0x02, East = 0x03, SouthEast = 0x04, South = 0x05, SouthWest = 0x06,
        West = 0x07, NorthWest = 0x08
    }

    /// <summary>
    /// Struct to handle coordinates
    /// </summary>
    /// 
    public struct RelativeCoordinates // relative coordinates
    {
        public int a;
        public int b;

        public RelativeCoordinates(int x, int y)
        {
            a = x;
            b = y;
        }
    }

    /// <summary>
    /// Class to coordinate all actions of all existing Settlers
    /// </summary>
    /// 
    sealed class SettlerAdmin
    {
        public int numSettlers = 1;
        public SettlerToDo[] XSettlerJob = new SettlerToDo[256];
        SettlerFunctions SettlerFunction = new SettlerFunctions();

        public SettlerAdmin()
        {
            for (int i = 0; i < 256; i++) XSettlerJob[i] = new AI.SettlerToDo();
        }

        public bool commandSettlerMove(Unit unit, RC RelCoor, Settler Settler, XCity[] XCities, int numCity, int sNum, GridMap XGridMap, Location[] CityLocs)
        {
            Location unitMoveTo;
            RC XDistVector = new RC(0, 0);
            double XDistValue = 0;

            XDistVector = unit.Location - XCities[numCity].cityLocation;
            XDistValue = XDistVector.Distance;

            if (XSettlerJob[sNum] == SettlerToDo.MoveToCity)
            {
                // If Settler was assigned to a City, now move to city location
                if (XDistValue <= 3) XSettlerJob[sNum] = SettlerToDo.BuildIrrigation;
                else Settler.moveToCity(unit, XCities[numCity], Settler, sNum);
            }

            if (XSettlerJob[sNum] == SettlerToDo.BuildIrrigation)
            {
                // The first task of the first settler will be to build irrigation
                // Therefore first look for Grassland, Prairie or Plains
                if (!(unit.Location == XCities[numCity].cityLocation) &&
                ((unit.Location.Terrain == Terrain.Prairie) ||
                (unit.Location.Terrain == Terrain.Plains) ||
                (unit.Location.Terrain == Terrain.Grassland)) &&
                (!(unit.Location.Improvement == TerrainImprovement.Irrigation)))
                {
                    unit.StartJob__Turn(Job.Irrigate);
                }
                else
                {
                    // Look around for Prairie, Plains or Grassland and move to it												
                    RelativeCoordinates RelativeCoor = new RelativeCoordinates(0, 0);
                    RelativeCoor = SettlerFunction.seekForIrrigationLocation(XCities[numCity].cityLocation);
                    if ((RelativeCoor.a == 0) && (RelativeCoor.b == 0)) XSettlerJob[sNum] = SettlerToDo.BuildRoads;
                    else
                    {
                        unitMoveTo = unit.Location + new RC(-RelCoor.a + RelativeCoor.a, -RelCoor.b + RelativeCoor.b);
                        unit.MoveTo__Turn(unitMoveTo);
                    }
                }
            }

            if (XSettlerJob[sNum] == SettlerToDo.BuildRoads)
            {
                // The second task of the first settler will be to build roads
                // Therefore first look for Grassland, Prairie or Plains with irrigation				
                if (!(unit.Location == XCities[numCity].cityLocation) &&
                   (unit.Location.Improvement == TerrainImprovement.Irrigation) &&
                   (!(unit.Location.HasRoad)) && (!(unit.Location.HasRiver)))
                {
                    unit.StartJob__Turn(Job.BuildRoad);
                }
                else
                {
                    // Look around for Prairie, Plains or Grassland and move to it													
                    RelativeCoordinates RelativeCoor = new RelativeCoordinates(0, 0);
                    
                    RelativeCoor = SettlerFunction.seekForRoadLocation(XCities[numCity].cityLocation);
                    if ((RelativeCoor.a == 0) && (RelativeCoor.b == 0))
                    {
                        XSettlerJob[sNum] = SettlerToDo.SeekNewSettleLocation;
                        Settler.seekNewSettleLocation(unit, XCities[numCity], Settler, sNum, XGridMap, CityLocs);
                        XCities[numCity].cityNeedsImprovements = false;
                    }
                    else
                    {
                        unitMoveTo = unit.Location + new RC(-RelCoor.a + RelativeCoor.a, -RelCoor.b + RelativeCoor.b);
                        unit.MoveTo__Turn(unitMoveTo);
                    }
                }
            }


            // If all irrigations and roads have been built it is time to seek a new Settling Site
            if (XSettlerJob[sNum] == SettlerToDo.SeekNewSettleLocation)
            {
                XCities[numCity].cityNeedsImprovements = false;
                Settler.seekNewSettleLocation(unit, XCities[numCity], Settler, sNum, XGridMap, CityLocs);
            }
            return false;
        }
    }


    /// <summary>
    /// Class to control a Settler and define routines for jobs and moves
    /// </summary>
    /// 
    sealed class Settler
    {
    	SettlerFunctions SettlerFunction = new SettlerFunctions();
    	
        public static int last_moveToGrid = 0;
        public int moveToGrid;
        public int uID;
        public bool exist;
        public bool unitIsOnGridLoc;
        public int movingCount;
        public bool isAssignedToCity;
        public bool settlerHasOrders;
        public bool settlerHasMoved;

        public Random random = new Random();

        public Settler(int ID, int sNum)
        {
            uID = ID;
            exist = true;
            unitIsOnGridLoc = false;
            movingCount = 0;
            isAssignedToCity = false;
            moveToGrid = sNum;
            settlerHasMoved = false;
            settlerHasOrders = false;
        }

        bool testSettleDown(Location unitLoc, XCity ACity, Location[] CityLoc)
        {
            bool canSettleDown = false;
            RC XDistVector = new RC(0, 0);
            double XDistValue = 0;
            XDistVector = unitLoc - CityLoc[0];
            XDistValue = XDistVector.Distance;

            if ((unitLoc.Terrain == Terrain.Grassland)
                || (unitLoc.Terrain == Terrain.Plains))
            {

                //Check Distance to home city
                if ((XDistValue >= 9) && (true) && (SettlerFunction.numberOfFoodLand(unitLoc)) >= 2) canSettleDown = true;
                for (int i = 0; i <= ACity.readNumOfCities(); i++)
                {
                    if (unitLoc == CityLoc[i]) canSettleDown = false;
                }

                // Check Distance to further cities
            }
            return canSettleDown;
        }



        public void seekNewSettleLocation(Unit unit, XCity XCities, Settler Settler, int sNum, GridMap XGridMap, Location[] CityLocs)
        {
            RelativeCoordinates RelCoor = new RelativeCoordinates(0, 0);
            bool settleDownOnThisLocation = false;

            // Here the location is tested for settling down
            settleDownOnThisLocation = testSettleDown(unit.Location, XCities, CityLocs);
            if (settleDownOnThisLocation)
            {
                unit.StartJob__Turn(CevoAILib.Job.BuildCity);
                exist = false;
            }

            // This part will move to a defined location, but only if the previous algorithm did not find a proper foodland place 
            else moveToGridMapLocation(unit, Settler, sNum, XGridMap);
        }

        public void moveToGridMapLocation(Unit unit, Settler Settler, int sNum, GridMap XGridMap)
        {
            RC RelCoor = new RC(0, 0);
            RelativeCoordinates RelativeCoor = new RelativeCoordinates(0, 0);
            Location unitMoveTo;

            RelCoor = XGridMap.GridMapEntity[moveToGrid].GridLocation - unit.Location;
            if ((RelCoor.a == 0) && (RelCoor.b == 0)) unitIsOnGridLoc = true;

            if (unitIsOnGridLoc)
            {
                RelativeCoor.a = random.Next(-1, 2);
                RelativeCoor.b = random.Next(-1, 2);
                unitMoveTo = unit.Location + new RC(RelativeCoor.a, RelativeCoor.b);
            }
            else
            {
                RelativeCoor = SettlerFunction.calculateRelCoors(RelCoor);
            }
                
            unitMoveTo = unit.Location + new RC(RelativeCoor.a, RelativeCoor.b);

            if ((unitMoveTo.Terrain == Terrain.Shore) || (unitMoveTo.Terrain == Terrain.Ocean))
            {
                do
                {
                    RelativeCoor.a = random.Next(-1, 2);
                    RelativeCoor.b = random.Next(-1, 2);
                    unitMoveTo = unit.Location + new RC(RelativeCoor.a, RelativeCoor.b);

                } while (((RelativeCoor.a == 0) && (RelativeCoor.b == 0)) ||
                         (unitMoveTo.Terrain == Terrain.Shore) ||
                         (unitMoveTo.Terrain == Terrain.Ocean));
            }

            unit.MoveTo__Turn(unitMoveTo);
            movingCount += 1;
        }

        public bool moveToCity(Unit unit, XCity XCities, Settler Settler, int sNum)
        {
            RC XDistVector = new RC(0, 0);
            double XDistValue = 0;

            XDistVector = unit.Location - XCities.cityLocation;
            XDistValue = XDistVector.Distance;

            RC RelCoor = new RC(0, 0);
            Location unitMoveTo;

            RelCoor = XCities.cityLocation - unit.Location;

            //Return true if Settler is near assigned City
            if (XDistValue <= 3) return true;

            // Move to City
            unitMoveTo = unit.Location + new RC(RelCoor.a, RelCoor.b);
            unit.MoveTo__Turn(unitMoveTo);
            return false;
        }
    }
}